# Copyright (c) 2021 Trelent Inc.

def get_python_indentation(lines, function_name, function_argument_count):
    """
    Returns the indentation level of a Python function's docstring.

    Parameters:

        lines (list): A list of strings, each representing a line from the file.

        function_name (str): The name of the function whose docstring indentation we are looking for.

        function_argument_count (num): The number of arguments this function contains.


    Generated by Trelent
    """

    # Constants
    def_str = "def " + function_name + "("
    end_str = "):"

    # Main loop
    i = 0
    start_index = 0
    check_multi = False
    while i < len(lines):

        # Check if we are in a multi-line definition
        if(check_multi is True):
            if(end_str in lines[i]):

                # Combine lines from start to end
                combined_definition = '\n'.join(lines[start_index:i])
                native_args = len(combined_definition.split(','))

                # Ensure we are within index bounds and are not in a duplicate function
                if(native_args == function_argument_count and i+1 < len(lines)):
                    return len(lines[i+1]) - len(lines[i+1].lstrip())
        else:
            # Look for single-line definition
            if(def_str in lines[i]):
                # Now check if we need more lines
                if(end_str in lines[i]):
                    
                    # Compute the number of arguments in this function
                    native_args = len(lines[i].split(','))

                    # Ensure we are within index bounds and are not in a duplicate function
                    if(native_args == function_argument_count and i+1 < len(lines)):
                        return len(lines[i+1]) - len(lines[i+1].lstrip())
                else:
                    # Set our multi-line flag
                    start_index = i
                    check_multi = True
        i += 1

    return -1

def get_javascript_indentation(lines, function_name):
    """
    Returns the indentation level of a Javascript function's docstring.

    Parameters:

        lines (list): A list of strings, each representing a line from the file.

        function_name (str): The name of the function whose docstring indentation we are looking for.


    Generated by Trelent
    """

    # Function signature
    def_str = "function " + function_name + "("

    # Main loop
    i = 0
    while i < len(lines):
        if(def_str in lines[i]):
            return len(lines[i]) - len(lines[i].lstrip())
        i += 1

    return -1

def get_indentation(lines, function_name, function_argument_count, language_id):
    if(language_id == "python"):
        return get_python_indentation(lines, function_name, function_argument_count)
    elif(language_id == "javascript"):
        return get_javascript_indentation(lines, function_name)
    #elif(language_id == "java"):
    #    return get_java_indentation(lines, function_name, function_argument_count)
    else:
        # Default to python
        return get_python_indentation(lines, function_name, function_argument_count)


# Copyright (c) 2021 Trelent Inc.

# External modules
from difflib import Differ
import math
import os
import requests
import sys
import textwrap

# Internal modules
import trelent.utils as utils


verbose = False

documented_function_counts = {
    "python": 0,
    "java": 0,
    "javascript": 0
}

num_documented_funcs = 0
num_skipped_funcs = 0
num_invalid_funcs = 0

def document_function(func, func_def, func_code, indentation, input_path, file_text, user_api_key):
    # TODO: func_def is based on what is parsed by the tree_hugger lib, which may not
    # always match the source code definition. As a result, if no match is found
    # the docstring will not be used. We need to do our own parsing eventually because
    # we will miss many functions this way. We log these in repo_metadata so that we can track our
    # progress over time: missed_funcs = num_functions - (num_originally_documented + num_trelent_documented)

    # Check for a definition mis-match
    if(func_def not in file_text):
        return file_text

    print("Generating for function '" + func + "' in '" + input_path + "'")
    # Call the Trelent API to get a docstring for the given function
    result = requests.post("https://trelent.npkn.net/document-public-repo/", json={
        "branded": False,
        "snippet": func_code,
        "language": "python",
        "sender": "trelent-cli"
    }, headers={
        "Api-Key": user_api_key,
        "X-Trelent-API-Key": user_api_key
    })

    # Create our docstring
    doc_str = result.json()["docstring"]
    doc_str = textwrap.indent(doc_str, ' '*indentation)
    new_func_def = func_def + doc_str + '\n'

    # Update our code!
    file_text = file_text.replace(func_def, new_func_def)
    documented_function_counts["python"] += 1
    return file_text

def document_folder(api_key, should_insert, folder, verbose_output):

    # Set whether or not to output verbose
    verbose = verbose_output

    # Recursively search the provided directory for source code files
    for root, dirs, files in os.walk(folder):
        for filename in files:

            # Check if this is a python file
            input_path = os.path.join(root, filename)
            lang = input_path[-3:]

            #  or lang == ".js" or lang == ".java"
            if(lang == ".py" and "test" not in input_path and "docs" not in input_path and "documentation" not in input_path and "examples" not in input_path):
                if(lang == ".py"):
                    lang = "python"
                    end_str = ":\n"
                elif(lang == ".js"):
                    lang = "javascript"
                    end_str = ") {"
                else:
                    # Default to Python
                    lang = "python" 
                    end_str = ":\n"

                # Document it!
                if(verbose): print("Parsing " + input_path + "...")

                with open(input_path, 'r+') as file:

                    # Split our file into lines
                    file_text = file.read()
                    original_text = file_text
                    lines = file_text.splitlines(keepends=True)

                    # Call our parsing API
                    parsed_file = requests.get("https://lambda.trelent.net/api/ParseFileFunctions", json={
                        "source": file_text,
                    })

                    parsed_file = parsed_file.json()
                    #print(parsed_file)

                    function_names = parsed_file["names"]
                    function_docstrings = parsed_file["docstrings"]
                    function_boddies = parsed_file["boddies"]

                    # Iterate through all the function names and retrieve the underlying code
                    for func in function_names:

                        # Check if there is already a docstring for this function
                        if(func not in function_docstrings.keys() and func in function_boddies.keys()):

                            # Fetch the desired code
                            func_code = function_boddies[func]
                            token_approx = math.ceil(len(func_code) * 1.1 / 4)

                            # Only generate for functions sufficiently long to be effective
                            if(token_approx > 50):
                                if(token_approx < 4000):

                                    # Check if the function indentation is valid (something we can work with)

                                    # Get the location of our docstring
                                    #if(lang == "javascript"):
                                    #    async_func_def = "async function " + func + "("
                                    #    func_def = "function " + func + "("
#
                                    #    docstring_index = file_text.find(async_func_def)
#
                                    #    if(docstring_index == -1):
                                    #        docstring_index = file_text.find(func_def)
                                    #    else:
                                    #        func_def = async_func_def
#
                                    #    # Retrieve indentation from the source text
                                    #    indentation = utils.get_indentation(lines, func, -1, lang)
#
                                    #    docstring_index -= (indentation + 1)
                                    #else:

                                    # Retrieve function code
                                    func_code = function_boddies[func]
                                    docstring_index = func_code.find(end_str) + 2
                                    func_def = func_code[0:docstring_index]

                                    if("," in func_def):
                                        num_args = len(func_def.split(','))
                                    else:
                                        num_args = 1

                                    # Retrieve indentation from the source text
                                    indentation = utils.get_indentation(lines, func, num_args, lang)
                            
                                    if(indentation != -1):
                                        file_text = document_function(func, func_def, func_code, indentation, input_path, file_text, api_key)
                                    else:
                                        if(verbose): print(f"Invalid function: {func}. Possibly a duplicate function name or other indentation error.")
                                else:
                                    if(verbose): print("Invalid function. Too many tokens (it's too long).")
                            else:
                                if(verbose): print("Invalid function. Too few tokens (it's too short).")

                    # Overwrite our file if there were changes
                    if(original_text != file_text):

                        # Check for insert mode first
                        if(should_insert):
                            file.seek(0)
                            file.write(file_text)
                            file.truncate()
                            if(verbose): print("Added docstrings into " + input_path)
                        else:
                            # Generate diff file
                            differ = Differ()

                            original_lines = original_text.splitlines(keepends=True)
                            updated_lines = file_text.splitlines(keepends=True)

                            diff_file_text = ''.join(differ.compare(original_lines, updated_lines))

                            diff_file_path = input_path + ".diff"
                            with open(diff_file_path, 'w+') as diff_file:
                                diff_file.seek(0)
                                diff_file.write(diff_file_text)
                                diff_file.truncate()

                            if(verbose): print("Added docstrings into " + diff_file_path)

                    else:
                        if(verbose): ("File fully documented, no changes necessary.")
            else:
                if(verbose): print("Skipping non-python, doc, test or example file '" + input_path + "'")

    num_documented_funcs = documented_function_counts["java"] + documented_function_counts["javascript"] + documented_function_counts["python"]


    print(" ")
    print("###########################")
    print(" ")
    print(f'Finished documenting {folder}!')
    print(f'Documented {num_documented_funcs} functions.')
    if(should_insert):
        print(f'Inserted documentation directly into source files.')
    else:
        print(f'Inserted documentation into diff files next to source.')
    print(" ")
    print("###########################")


# Copyright (c) 2021 Trelent Inc.

# External modules
import configparser
import os
from pathlib import Path
import requests

# Setup our config file paths
home_path = str(Path.home())
config_dir = home_path + '/.trelent'
config_path = config_dir + '/config.ini'

# Create a default config file if one doesn't exist already
if not os.path.isfile(config_path):
    os.makedirs(config_dir, exist_ok=True)
    f = open(config_path, "w")
    f.write("[CREDENTIALS]\nTrelentAPIKey = none")
    f.close()

# Initialize our ConfigParser
config = configparser.ConfigParser()
config.read(config_path)

def get_api_key():
    """
    Returns the API key for Trelent.

    :returns: The API key for Trelent. If no API key is found, returns None.
    """

    api_key = config["CREDENTIALS"]["TrelentAPIKey"]
    if(api_key == "none"):
        return None
    
    return api_key


def set_api_key(api_key):
    """
    Set the Trelent API key.

    :param api_key: The API key to set. Must be a string starting with "sk_trlnt_".
    :returns bool: True if the operation was successful, False otherwise.
    """
    if(api_key[:9] != "sk_trlnt_"):
        return False
    else:
        config["CREDENTIALS"]["TrelentAPIKey"] = api_key
        with open(config_path, 'w+') as configfile:
            config.write(configfile)
        return True

def is_valid_api_key(api_key):
    result = requests.post("https://trelent.npkn.net/check-api-key/", headers={
        "X-Trelent-Api-Key": api_key
    })

    return result.status_code == 200


# Copyright (c) 2021 Trelent Inc.

# External modules
import click

# Internal modules
import trelent.auth_module as auth_module
import trelent.docgen as docgen

@click.group()
def trelent():
    pass

@click.command()
@click.password_option(prompt='Your API Key')
def auth(password):
    if auth_module.set_api_key(password) and auth_module.is_valid_api_key(password):
        click.echo(u'\u001b[32;1mAdded API Key Successfully!')
    else:
        click.echo(u'\u001b[31;1mThe API Key you used is invalid. Please copy your API Key from our dashboard.')

@click.command()
@click.option('--key', '-k', help="Provide an API Key directly without storing in plaintext within a config file.")
@click.option('--verbose', '-v', is_flag=True, help="Should we print verbose logs?")
@click.option('--insert', '-i', is_flag=True, help="Should we automatically insert the docstrings?")
@click.argument('folder')
def document(folder, insert, verbose, key):
    click.echo(u'\u001b[33;1mAttempting to document python files in ' + folder + u'\u001b[33;1m  ...')

    # Check for authentication
    if(key != None):
        # Use provided api key
        api_key = key

    else:
        # Check config file
        api_key = auth_module.get_api_key()
        if(api_key == None):
            click.echo(u"\u001b[31;1mYou have not added your API Key yet! Try running 'trelent auth' or provide your api key directly using the '--key' option")
            return

    if(not auth_module.is_valid_api_key(api_key)):
        click.echo(u"\u001b[31;1mThe API Key you used is invalid. Please copy your API Key from our dashboard.")
        return

    try:
        docgen.document_folder(api_key, insert, folder, verbose)
        click.echo(u'\u001b[32;1mSuccessfully documented ' + folder + u'\u001b[32;1m!')
    except Exception as e:
        click.echo(u'\u001b[31;1mError: ' + str(e))
        click.echo(u'\u001b[31;1mAn error occured while documenting the provided folder')

trelent.add_command(auth)
trelent.add_command(document)

if __name__ == '__main__':
    trelent()


# Copyright (c) 2021 Trelent Inc.

# External modules
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="trelent",
    author="Trelent Inc.",
    author_email="contact@trelent.net",
    description="A command line tool to generate documentation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://trelent.net",
    version='1.0',
    py_modules=['trelent.auth_module', 'trelent.docgen', 'trelent.main', 'trelent.utils'],
    install_requires=[
        'Click',
        'Requests'
    ],
    entry_points='''
        [console_scripts]
        trelent=trelent.main:trelent
    ''',
    classifiers=[
        "Operating System :: OS Independent",
    ],
)